#ifndef __Assembler__
#include	<Assembler.h>
#endif

#include	"TrapAvailable.h"
/*
; CDIconKiller  1994-1997 Quinn & Peter N Lewis, Fabrizio Oddone
; Version 1.0 written in 4 hours from 02:00 to 05:30 18/June/94

bug fixes, CD driver string in resource (suggested by Quinn),
asm tweaks, cdev trickery, etc.
(version 1.1 thru 1.5) by Fabrizio Oddone

version 1.5 updated in 3 hours from 16:00 to 19:00 (CET) October 17, 1997
dreaded "File Sharing startup crash" fixed from 10:00 to 11:00 October 18, 1997
bottled wine (Barbera) in the afternoon -- BTW, check out the 1997 Italian wines: a most excellent crop this year
LoadADrive compatibility reassessment & code tweaks from 15:00 to 17:00 October 19, 1997

; Patches GetCatInfo, tests for Custom Icon attribute on
; CD-ROM files/folders, and clears it.

; Detailed description:

; now runs as a cdev so the patch affects only the Finder
; since there is no user interface, we can retain the old icon

; macDev time:
;	check if we are running under System 7 or later
;		check if the patch is already installed
;		load the resource containing the driver strings
;		patch GetCatInfo

; initDev time:
;	we are done, quit silently

; GetCatInfo patch (simplified):
;	Test for GetCatInfo selector on _HFSDispatch, do nothing otherwise
;	Test for async call; if so, patch completion routine
;	Call thru
;	Return immediately on error (async calls are handled by the completion patch)
;	Return immediately unless Custom Icon bit set
;	Walk the volume queue looking for the vRefNum of the GetCatInfo
;	If we don't find it, we leave
;	If the command key is down, we leave
;	We determine whether the driver refnum has been seen already
;	If yes
;		determine if the custom icon bit must be cleared for this driver
;	If no
;		we scan the driver queue and update our lists

Very brief discussion about the async patch:
We cannot simply tail patch the call (being async, it hasn't run yet);
we can tail patch the completion routine, though.
We don't handle the worst case scenario with many outstanding async calls, that is,
many completion routine addresses to save. We only save ONE completion routine address.
Basically our patch can be in two states: can/cannot handle an async call, i.e.
can/cannot save the completion routine address.
The state indicator is bit #0 in the completion routine address
(routine addresses must be even, that's why this bit can be used).
Moreover, we can use BCLR to atomically test and clear the bit,
in order to eschew any foreseeable reentrancy problem.

	include	'sysequ.a'
	include	'fsequ.a'
	include	'traps.a'

; procedure ShowIcon (id: integer; delta_x: integer);
	import	ShowIcon
main	proc	export

extern pascal void ShowIcon(short iconID, short pixels);
*/
static void mypatch(void);

//#define patchSize	(EndHFSDispatch - OurHFSDispatch)
#define	patchSize	(650)	// because of CodeWarrior asm limitations
#define	HFSDispatchTrapNum	0x60
#define	kNumKilledDrivers	15
#define	kNumOtherDrivers	63

pascal long main(short message, short /*item*/, short /*numItems*/, short /*cPanelID*/, EventRecord */*theEvent*/, long cdevValue, DialogPtr /*cpDialog*/)
{
long	resp;

if (cdevValue) {
	switch (message) {
		case initDev:
			return cdevGenErr;
			break;
		case macDev:
// normally you shouldn't check against the System version but against individual features
// but in this case we want to know whether custom icons are used by the Finder
			if (TrapAvailable(_Gestalt)) {
				if (Gestalt(gestaltSystemVersion, &resp) == noErr)
					if (resp >= 0x700) {
						mypatch();
						return 1;
						}
				}
// no _Gestalt => very old System (< 6.0.4)
			return 0;	
			break;
		}
	}
return cdevValue;
}

asm void mypatch(void)
{
	moveq	#HFSDispatchTrapNum,d0	// patch HFSDispatch
	_GetOSTrapAddress
	move.l	OurHFSDispatch,d0
	cmp.l	(a0),d0
// if the patch is already there do nothing
	beq.s	bad
	lea	oldHFSDispatch,a1
	move.l	a0,(a1)

	subq.w	#4,sp
	move.l	#'STR#',-(SP)
	move.w	#128,-(sp)
	_Get1Resource
	move.l	(sp)+,d0
	beq.s	bad
	lea	stringH,a0
	move.l	d0,(a0)	// remember string handle
	move.l	d0,-(sp)
	_DetachResource
	
	bsr	findDriverbyName

	move.l	#patchSize,d0
	_NewPtr	//Sys
// this block is installed in the System heap anyway, because
// the Finder calls our cdev with the current zone set to the System zone
// (at least under 7.5.5!)
// in theory we should install everything in the Finder heap since it now
// belongs there; but since the Finder malfunctions when its heap is low
// (maybe the current zone is not set to the Finder zone for this reason)
// we refrain from forcing the Finder zone with GetZone/SetZone
	bne.s	baddisposepref

	movea.l	a0,a1
	lea	OurHFSDispatch,a0
	move.l	#patchSize,d0
	_BlockMove

	moveq	#HFSDispatchTrapNum,d0
	movea.l	a1,a0
	_SetOSTrapAddress	

//	MOVE.L	#0xffff0080,-(SP)
//	JSR	ShowIcon
bad:
	rts

baddisposepref:
	movea.l	stringH,a0
	_DisposHandle
	rts

// BEGIN may run at interrupt time
OurHFSDispatch:
	cmp.w	#9,d0	// Test for GetCatInfo selector
	bne.s	dontdoit

// the trap dispatcher always puts the trapword in register D1
	btst	#10,d1	// sync?
	beq.s	tailpatch	// tail patch only sync calls

// for async calls, we intercept the completion routine
	move.l	a1,-(sp)
	lea		savedCompletion+3,a1
	bclr	#0,(a1)	// test and clear -- be reentrant!
	beq.s	restoreA1fin	// cannot save over existing completion

	subq.w	#3,a1	// point to savedCompletion
	move.l	struct (DirInfo.ioCompletion)(a0),(a1)	// save
	lea		myCompletion,a1
	move.l	a1, struct (DirInfo.ioCompletion)(a0)	// put our own
restoreA1fin:
	movea.l	(sp)+,a1
	bra.s		dontdoit

tailpatch:
	pea	continued
dontdoit:
	move.l	oldHFSDispatch,-(sp)	// Call thru
	rts
// END may run at interrupt time

continued:
	tst.w	struct (DirInfo.ioResult)(a0)
	bne.s	fin	// give up on error

	btst	#2,struct (DirInfo.ioDrUsrWds.frFlags)(a0)
	beq.s	fin	// give up unless custom icon
	
	movem.l	a1/d1,-(sp)
	
	bsr		cmdkeyisdown
	bne.s	leave

	movea.l	0x358,a1	// first volume control block LMGetVCBQHdr()+2
	move.w	struct (DirInfo.ioVRefNum)(a0),d1
lop:
	cmp.w	struct (VCB.vcbVRefNum)(a1),d1
	bne.s	cont

	move.w	struct (VCB.vcbDRefNum)(a1),d1
	bsr	finddriverdontwantcustomicons
	bne.s	clearcustombit
	bsr	finddriverwantcustomicons
	bne.s	leave
	bsr	findDriverbyName
	bra.s	leave
clearcustombit:
	bclr	#2,struct (DirInfo.ioDrUsrWds.frFlags)(a0)	// clear custom icon
	bra.s	leave

cont:
	movea.l	(a1),a1	// next volume control block
	move.l	a1,d0
	bne.s	lop

leave:
	movem.l	(sp)+,a1/d1
	moveq	#0,d0
	
fin:
	tst.w	d0
notstfin:
	rts

// BEGIN may run at interrupt time
cmdkeyisdown:
	MOVE.L	A0,-(SP)
	LEA	-sizeof(EventRecord)(SP),SP
	MOVEA.L	SP,A0
	MOVEQ	#0,D0
	_OSEventAvail	// can be called at interrupt time; cool isn't it?
	BTST	#0,struct (EventRecord.modifiers)(A0)	//cmdKey
	LEA	sizeof(EventRecord)(SP),SP
	MOVEA.L	(SP)+,A0
	RTS

myCompletion:
	tst.w	struct (DirInfo.ioResult)(a0)
	bne.s	endcompletion	// give up on error

	btst	#2,struct (DirInfo.ioDrUsrWds.frFlags)(a0)
	beq.s	endcompletion	// give up unless custom icon
	
	bsr		cmdkeyisdown
	bne.s	endcompletion

	movea.l	0x358,a1	// first volume control block LMGetVCBQHdr()+2
	move.w	struct (DirInfo.ioVRefNum)(a0),d1
comploop:
	cmp.w	struct (VCB.vcbVRefNum)(a1),d1
	bne.s	compcont

	move.w	struct (VCB.vcbDRefNum)(a1),d1
	bsr.s	finddriverdontwantcustomicons
	beq.s	endcompletion
	bclr	#2,struct (DirInfo.ioDrUsrWds.frFlags)(a0)	// clear custom icon
	bra.s	endcompletion
compcont:
	movea.l	(a1),a1	// next volume control block
	move.l	a1,d0
	bne.s	comploop

endcompletion:
	movea.l	savedCompletion,a1
	move.l	a1,struct (DirInfo.ioCompletion)(a0)	// without this line, File Sharing will crash when starting up!
	beq.s	nilcompletion
	jsr	(a1)
nilcompletion:
	lea	savedCompletion+3,a1
	bset	#0,(a1)	// can reuse for next async call
	
	rts

finddriverdontwantcustomicons:
	move.l	a0,-(sp)
	lea	lastskipthesedrivers,a0
	bset	#7,(a0)	// eschew reentrancy problems; all drivers should have negative refnum
	bne.s	alreadyinitiated
	move.w	d1,(a0)
	lea	skipthesedrivers,a0
loopfinddontwant:
	cmp.w	(A0)+,D1
	bne.s	loopfinddontwant

	move.l	a0,-(SP)
	lea	lastskipthesedrivers,a0
	bclr	#7,(a0)	// clear
	movea.l	(sp)+,a0

alreadyinitiated:
	pea	lastskipthesedrivers+2
	cmpa.l	(SP)+,A0
	movea.l	(sp)+,a0
	rts
// END may run at interrupt time

finddriverwantcustomicons:
	move.l	a0,-(sp)
	lea	lastotherdrivers,a0
	move.w	D1,(a0)
	lea	otherdrivers,a0
loopfindwant:
	cmp.w	(A0)+,D1
	bne.s	loopfindwant
	pea	lastotherdrivers+2
	cmpa.l	(sp)+,A0
	movea.l	(sp)+,a0
	rts

findDriverbyName:
	movem.l	a0-a4/a6/d1/d3-d5,-(sp)
	moveq	#0,d5
	lea	nextfreedontwant,a3
	tst.w	(a3)
	bge.s	alreadyinitednfdw
	st	d5
	clr.w	(a3)
alreadyinitednfdw:
	moveq	#(kNumKilledDrivers*2),d0
	cmp.w	(a3),d0
	beq.s	plainexit

	lea	nextfreewant,a4
	moveq	#(kNumOtherDrivers*2),d0
	cmp.w	(a4),d0
	beq.s	plainexit

	movea.l	0x11C,a2	// LMGetUTableBase()
	move.w	0x1d2,d3	// LMGetUnitTableEntryCount()
loop:
	move.l	(a2)+,d0
	beq.s	endloop
	movea.l	d0,a1

// a1 = curDCtlHndl

	movea.l	(a1),a1
	move.l	/*struct (DCtlEntry.dCtlDriver)*/(a1),d0
	beq.s	endloop
	movea.l	d0,a0
	btst	#6,(struct (DCtlEntry.dCtlFlags)+1)(a1)	// test for RAMDriver
	beq.s	notram
	move.l	(a0),d0
	beq.s	endloop
	movea.l	d0,a0
notram:
	move.w	d3,d4
	sub.w	0x1d2,d4	// calc driver refnum -(unit num+1) LMGetUnitTableEntryCount()
	subq.w	#1,d4
	tst.b	d5
	bne.s	firsttime
	move.w	d4,d1
	bsr	finddriverdontwantcustomicons
	bne.s	endloop
	bsr	finddriverwantcustomicons
	bne.s	endloop
firsttime:
	lea	struct (DRVRHeader.drvrName)(a0),a0
	moveq	#0,d2
	move.b	(a0)+,d2
	swap	d2
	bsr	cmpallstrings
	bne.s	endloopsavedriver

	lea	skipthesedrivers,a0
	movea.l	a3,a1
	bra.s	commonpart
endloopsavedriver:
	lea	otherdrivers,a0
	movea.l	a4,a1
commonpart:
	adda.w	(a1),a0
	move.w	d4,(a0)	// remember driver refnum
	addq.w	#2,(a1)
endloop:
	subq.w	#1,d3
	bne	loop
plainexit:
	movem.l	(sp)+,a0-a4/a6/d1/d3-d5
	rts

// input: a0 string, d2 (already in hi word) length
// output: sets the flags
// registers are saved&restored by the caller
cmpallstrings:
	movea.l	stringH,a6
	movea.l	(a6),a6
	move.w	(a6)+,d1
	andi	#0xB,CCR
	bra.s	thedbcc
cmploop:
	move.l	d2,d0
	move.b	(a6)+,d0
	movea.l	a6,a1
	adda.w	d0,a6
	_CmpString
thedbcc:
	dbeq	d1,cmploop
cmpexit:
	rts

// list of drivers whose disks won't display custom icons
skipthesedrivers:
	ds.w	kNumKilledDrivers
lastskipthesedrivers:
	dc.w	1
// list of the remaining drivers
otherdrivers:
	ds.w	kNumOtherDrivers
lastotherdrivers:
	dc.w	1

nextfreedontwant:
	dc.w	-1
nextfreewant:
	dc.w	0
stringH:
	dc.l	0
savedCompletion:
	dc.l	1
oldHFSDispatch:
	dc.l	0xffffffff
EndHFSDispatch:

//	string	pascal

}
//	endp
	
//	end
	
/*
asm -wb "{active}"
(evaluate "{active}" =~ "/(?*:)1(?*)/" )> dev:null
(evaluate "{active}" =~ "/(?*)2.a/" )> dev:null
link -rt INIT=128 -ra =ressysheap,reslocked -t INIT -c CDik -o "{2}" "{active}.o" "{1}ShowIcon.o"
setfile -a B "{2}"



dumpobj "{1}SHOWICON.o"
*/
